import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart' as storage;
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';

/// Accessing the instantiated `AuthApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `AuthApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<AuthApi>().getCurrentUser();
/// ```
///
abstract class AuthApi {
  Future<AuthToken> getAuthTokenFromCache();

  Future<void> cacheAuthToken(AuthToken authToken);

  /// Returns m4e user with `user-id`
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<User> getM4eUser(UniqueId userId);

  /// Returns m4e access-token with `refresh-token`
  ///
  /// Throws `ServerException` when server call fails or refresh token is invalid
  Future<AuthToken> refreshAccessToken(AuthToken authToken);

  /// Returns `user-id` of phone-number sent an OTP
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  Future<UniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm);

  // TODO: implement
  Future<AuthToken> verifyOTPCode(
      M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm);
}

class AuthApiImpl implements AuthApi {
  final Dio _client;
  final storage.FlutterSecureStorage _storage;

  AuthApiImpl({
    @required Dio client,
    @required storage.FlutterSecureStorage storage,
    @required M4eLoggingInterceptors loggingInterceptor,
  })  : assert(client != null),
        assert(storage != null),
        assert(loggingInterceptor != null),
        _client = client,
        _storage = storage {
    // logs requests
    _client?.interceptors?.add(loggingInterceptor);
  }

  String _base64EncodedChallenge(
      M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm) {
    final _headerParams =
        '${m4eAuthOTPVerificationForm.id}:${m4eAuthOTPVerificationForm.otpCode}:${DateTime.now().toTimestampSinceEpoch()}';
    return base64.encode(_headerParams.codeUnits);
  }

  Future<AuthToken> getAuthTokenFromCache() async {
    final _authTokenString =
        await _storage.read(key: Persistence.kM4eAuthToken);

    if (_authTokenString != null) {
      return AuthToken.fromJson(json.decode(_authTokenString));
    }

    throw ExceptionMessages.kNoAuthUser;
  }

  Future<void> cacheAuthToken(AuthToken authToken) async {
    final _authTokenString = json.encode(authToken.toJson());

    return await _storage.write(
        key: Persistence.kM4eAuthToken, value: _authTokenString);
  }

  // REMOTE===========================================================================================

  Future<User> getM4eUser(UniqueId userId) async {
    // TODO: implement fetching user with user-id
    return Future.value(null);
  }

   Future<AuthToken> refreshAccessToken(AuthToken authToken) async {
     // TODO: implement refresh-token scheme
     return authToken;
   }

  Future<UniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm) async {
      assert(m4eAuthPhoneNumberForm != null,
          'M4eAuthPhoneNumberForm should not be null');
    try {

      final _response = await _client.put('/auth/claim',
          data: json.encode(m4eAuthPhoneNumberForm.toJson()));

      if (_response.statusCode == 200) {
        return UniqueId.fromJson(_response.data);
      }

      throw ExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw ExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw ExceptionMessages.kUndefinedException;
    } on SocketException {
      throw ExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw ExceptionMessages.kUndefinedException;
    }
  }

  Future<AuthToken> verifyOTPCode(
      M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm) async {
      assert(m4eAuthOTPVerificationForm != null,
          'M4eAuthOTPVerificationForm should not be null');
    try {

      final _response = await _client.put(
        '/auth/verify',
        data: json.encode(m4eAuthOTPVerificationForm.toJson()),
        options: Options(
          headers: {
            HttpHeaders.authorizationHeader:
                'Challenge ${_base64EncodedChallenge(m4eAuthOTPVerificationForm)}'
          },
        ),
      );

      if (_response.statusCode == 200) {
        return AuthToken.fromJson(_response.data);
      }

      throw ExceptionMessages.kFailedTransactionException;
    } on DioError catch (e) {
      // print error in console in debug mode
      logger(e);

      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw ExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw ExceptionMessages.kUndefinedException;
    } on SocketException {
      throw ExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw ExceptionMessages.kUndefinedException;
    }
  }
}
