import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';

/// Accessing the instantiated `WalletApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `WalletApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<WalletApi>().getSavedWalletsFromCache(userId);
/// ```
///
abstract class WalletApi {
  /// Returns list of wallets for current authenticated
  /// m4e user with `OWNER` role entity-wallet attributes
  Future<List<Wallet>> fetchUserWallets(UniqueId userId);

  /// Returns list of m4e wallets for user with id [userId]
  /// with `PAY` and `SEE` access permissions
  Future<List<Wallet>> fetchWalletsWithUserId(UniqueId userId);
}

class WalletApiImpl implements WalletApi {
  final Dio _client;

  WalletApiImpl({
    @required Dio client,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client {
    // logs and appends authorization headers to requests
    _client.interceptors.add(loggingInterceptor);
    _client.interceptors.add(headerInterceptor);
  }

  /// Returns list of wallets for current authenticated
  /// m4e user with `OWNER` role entity-wallet attributes
  Future<List<Wallet>> fetchUserWallets(UniqueId userId) async {
    try {
      final _response = await _client.get(
          '/wallets/$userId/access?access=SEE&access=VIEW&access=PAY&access=ALTER&access=GRANT&access=CREATE&access=DESTROY');

      if (_response.statusCode == 200) {
        final _walletsMap = _response.data['wallets'] as List;

        return _walletsMap?.map((e) => Wallet.fromJson(e))?.toList() ?? [];
      }

      throw ExceptionMessages.kUndefinedException;
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw ExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw ExceptionMessages.kUndefinedException;
    } on SocketException {
      throw ExceptionMessages.kNoInternetConnectionException;
    } catch (e) {
      throw ExceptionMessages.kUndefinedException;
    }
  }

  /// Returns list of m4e wallets for user with id [userId]
  /// with `PAY` and `SEE` access permissions
  Future<List<Wallet>> fetchWalletsWithUserId(UniqueId userId) async {
    try {
      // TODO: re-implement to fetch wallets current m4e user has PAY and SEE access to
      final _response =
          await _client.get('/wallets/$userId/access?access=PAY_TO&access=SEE');

      if (_response.statusCode == 200) {
        final _walletsMap = _response.data['wallets'] as List;

        return _walletsMap?.map((e) => Wallet.fromJson(e))?.toList() ?? [];
      }

      throw ExceptionMessages.kUndefinedException;
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw ExceptionMessages.kConnectionTimeOutException;
      }
      if (e.type == DioErrorType.RESPONSE) {
        if (e.response.statusCode == 401) {
          // TODO: implement mapping and returning error responses from server
          throw UnimplementedError();
        }
      }

      throw ExceptionMessages.kUndefinedException;
    } on SocketException {
      throw ExceptionMessages.kNoInternetConnectionException;
    } on AuthException catch (e) {
      rethrow;
    } catch (e) {
      throw ExceptionMessages.kUndefinedException;
    }
  }
}
