import 'package:data_connection_checker/data_connection_checker.dart';

import '../../m4e_sdk_initialize.dart';
import '../api/apis.dart';
import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../models/models.dart';

/// Exposes SDK Auth Api methods
///
/// Accessing the instantiated `m4eAuth` methods
/// is done by:
///
/// ```
/// final _m4eAuth = M4eAuth();
///
/// final _isAuthenticated = await _m4eAuth.isAuthenticated();
/// ```
///
class M4eAuth {
  final AuthApi _authApi;
  final DataConnectionChecker _connectionChecker;

  M4eAuth({
    AuthApi authApi,
    DataConnectionChecker connectionChecker,
  })  : _authApi = authApi ?? m4eSDK<AuthApi>(),
        _connectionChecker =
            connectionChecker ?? m4eSDK<DataConnectionChecker>();

  /// Returns `AuthToken` if user is authenticated
  ///
  /// Throws `AuthException` with `NO_AUTH_USER` error
  /// code if no user is authenticated
  ///
  Future<AuthToken> isAuthenticated() async {
    try {
      final _token = await _authApi.getAuthTokenFromCache();

      return _token;
    } on AuthException catch (e) {
      rethrow;
    }
  }

  /// Requests `access-token` from m4e using existing `refresh-token`
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  ///
  /// code device is not connected to the internet
  ///
  Future<AuthToken> refreshAccessToken() async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _existingToken = await _authApi.getAuthTokenFromCache();

        final _token = await _authApi.refreshAccessToken(_existingToken);

        // cache new token
        await _authApi.cacheAuthToken(_token);

        return _token;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }

  /// Returns an m4e user with `user-id`
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  ///
  /// code device is not connected to the internet
  ///
  Future<User> getM4eUser(UniqueId userId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
         // TODO: implement fetching user with user-id
        return Future.value(null);
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }

  /// Takes in `M4eAuthPhoneNumberForm` parameter to send
  /// verification-code (OTP) from m4e
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  ///
  /// code if device is not connected to the internet
  ///
  Future<UniqueId> verifyPhoneNumber(
      M4eAuthPhoneNumberForm m4eAuthPhoneNumberForm) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _userId =
            await _authApi.verifyPhoneNumber(m4eAuthPhoneNumberForm);

        return _userId;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }

  /// Takes in `M4eAuthOTPVerificationForm` parameter
  /// and returns `AuthToken` on success call
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code if device is not connected to the internet
  ///
  Future<AuthToken> verifyOTPCode(
      M4eAuthOTPVerificationForm m4eAuthOTPVerificationForm) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _authToken =
            await _authApi.verifyOTPCode(m4eAuthOTPVerificationForm);

        await _authApi.cacheAuthToken(_authToken);
        return _authToken;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }
}
