import 'package:data_connection_checker/data_connection_checker.dart';

import '../../core/core.dart';
import '../../m4e_forms/m4e_forms.dart';
import '../../m4e_sdk_initialize.dart';
import '../../models/models.dart';
import '../api/apis.dart';

/// Exposes SDK Transaction Api methods
///
/// Accessing the instantiated `m4eTransaction` methods
/// is done by:
///
/// ```
/// final _m4eTransaction = M4eTransaction();
///
/// await _m4eTransaction.makeTransaction(Transaction transaction);
/// ```
///
class M4eTransaction {
  final TransactionApi _transactionApi;
  final DataConnectionChecker _connectionChecker;

  M4eTransaction({
    TransactionApi transactionApi,
    DataConnectionChecker connectionChecker,
  })  : _transactionApi = transactionApi ?? m4eSDK<TransactionApi>(),
        _connectionChecker =
            connectionChecker ?? m4eSDK<DataConnectionChecker>();

  /// Requires `Transaction` form as parameter
  ///
  /// Throws [AuthException] if no m4e user is authenticated
  ///
  /// Throws [ServerException] if request fails. This could be as a result of a
  /// `Request timeout`, `Socket Exception` or `General server error`
  ///
  // TODO: Pass-in token from application when issueing a request
  Future<UnitImpl> makeTransaction(M4eTransactionForm transaction) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        await _transactionApi.makeTransaction(transaction);

        return UnitImpl();
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }

  /// Returns `TransactionLogList` of authenticated m4e user.
  ///
  /// Throws [AuthException] if no m4e user is authenticated
  ///
  /// Throws [CacheException] if no `TransactionLog` is returned from cache on no-internet connectivity
  ///
  /// Throws [ServerException] if request fails. This could be as a result of a
  /// `Request timeout`, `Socket Exception` or `General server error`
  ///
  // TODO: Pass-in token from application when issueing a request
  Future<TransactionLogList> getUserTransactionLogs(UniqueId userId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _transactionLogs =
            await _transactionApi.getUserTransactionLogs(userId);

        return _transactionLogs;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }

  /// Returns `TransactionLogList` of wallet whose wallet-id is passed as method parameter.
  ///
  /// Throws [AuthException] if no m4e user is authenticated
  ///
  /// Throws [CacheException] if no `TransactionLog` is returned from cache on no-internet connectivity
  ///
  /// Throws [ServerException] if request fails. This could be as a result of a
  /// `Request timeout`, `Socket Exception` or `General server error`
  ///
  // TODO: Pass-in token from application when issueing a request
  Future<TransactionLogList> getWalletTransactionLogs(UniqueId walletId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _transactionLogs =
            await _transactionApi.getWalletTransactionLogs(walletId);

        return _transactionLogs;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }
}
