import 'package:data_connection_checker/data_connection_checker.dart';
import 'package:uuid_type/uuid_type.dart';

import '../../m4e_sdk_initialize.dart';
import '../api/apis.dart';
import '../../models/models.dart';
import '../../core/core.dart';

/// Exposes SDK Wallet Api methods
///
/// Accessing the instantiated `m4eWallet` methods
/// is done by:
///
/// Instantiating as a regular object
///
/// ```
/// final _m4eWallet = M4eWallet();
///
/// await _m4eWallet.getUserWallets();
/// ```
///
class M4eWallet {
  final WalletApi _walletApi;
  final DataConnectionChecker _connectionChecker;

  M4eWallet({
    WalletApi walletApi,
    DataConnectionChecker connectionChecker,
  })  : _walletApi = walletApi ?? m4eSDK<WalletApi>(),
        _connectionChecker =
            connectionChecker ?? m4eSDK<DataConnectionChecker>();

  /// Returns current m4e authenticated user `Wallets` from m4e when
  /// connected to the internet or local cache (if exist) on no-internet-connectivity
  ///
  /// Throws [AuthException] if no m4e user is authenticated
  ///
  /// Throws [CacheException] if no `Wallets` are returned from cache on no-internet connectivity
  ///
  /// Throws [ServerException] if request fails. This could be as a result of a
  /// `Request timeout`, `Socket Exception` or `General server error`
  ///
  Future<List<Wallet>> getUserWallets(UniqueId userId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _wallets = await _walletApi.fetchUserWallets(userId);

        return _wallets;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }

  /// Returns m4e user `Wallets` with specified `user-id` from m4e when
  /// connected to the internet or local cache (if exist) on no-internet-connectivity
  ///
  /// Throws [AuthException] if no m4e user is authenticated
  ///
  /// Throws [CacheException] if no `Wallets` are returned from cache on no-internet connectivity
  ///
  /// Throws [ServerException] if request fails. This could be as a result of a
  /// `Request timeout`, `Socket Exception` or `General server error`
  ///
  /// Pass-in an m4e user-id to fetch his wallets
  Future<List<Wallet>> getWalletsWithUserId(UniqueId userId) async {
    if ((await _connectionChecker.hasConnection) ?? false) {
      try {
        final _wallets = await _walletApi.fetchWalletsWithUserId(userId);

        return _wallets;
      } catch (e) {
        rethrow;
      }
    }

    throw ExceptionMessages.kNoInternetConnectionException;
  }
}
